@@ -5,18 +5,19 @@
 #  ALL RIGHTS RESERVED.
 #
 #  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
+import asyncio
 import sys
 import traceback
 
 from run_markdown import get_logic_block_from_md_file
 
 
-def main():
+async def main():
     assert len(sys.argv) > 1
     target_file = sys.argv[1]
     block = get_logic_block_from_md_file(target_file)
     try:
-        block.evaluate()
+        await block.evaluate()
     except Exception:
         print("CHECK FAILED")
         traceback.print_exc()
@@ -25,4 +26,4 @@ def main():
 
 
 if __name__ == '__main__':
-    main()
+    asyncio.run(main())
