@@ -5,9 +5,8 @@
 #  ALL RIGHTS RESERVED.
 #
 #  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
+import asyncio
 import json
-import subprocess
-import traceback
 from dataclasses import dataclass, field
 from typing import Literal, Any
 
@@ -16,38 +15,73 @@ import mistune
 
 @dataclass
 class RunResult:
-    stdout: str
-    stderr: str
-    returncode: int
+    stdout: str = ''
+    process: asyncio.subprocess.Process | None = None
 
 
-def execute_code_block(code: str, language: str, timeout: float = 1) -> RunResult | subprocess.Popen:
+async def execute_code_block(code: str, language: str, timeout: float = 1) -> RunResult:
     if language == 'python':
-        process = subprocess.Popen(['python', '-c', code], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
+        process_args = ['python', '-c', code]
     elif language in ('shell', 'bash', 'sh'):
-        process = subprocess.Popen(code, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, shell=True)
+        process_args = ['bash', '-c', code]
     else:
         raise ValueError('Unknown language')
 
+    process: asyncio.subprocess.Process = await asyncio.create_subprocess_exec(
+        *process_args, stdout=asyncio.subprocess.PIPE, stdin=asyncio.subprocess.PIPE,
+        stderr=asyncio.subprocess.PIPE)
+
+    result = RunResult()
     try:
-        stdout, stderr = process.communicate(timeout=timeout)
-        return RunResult(stdout=stdout, stderr=stderr, returncode=process.returncode)
-    except subprocess.TimeoutExpired:
-        return process
-    except Exception:
-        traceback.print_exc()
-        raise
+        # await asyncio.wait_for(process.communicate(code.encode()), timeout=timeout)
+        await asyncio.wait_for(process.wait(), timeout=timeout)
+        # print('proc finished')
+    except asyncio.TimeoutError:
+        # print('proc timout')
+        result.process = process
+    stdout = b''
+    try:
+        while True:
+            red = await asyncio.wait_for(process.stdout.readline(), timeout=0.1)
+            # print(f'readline {red=}')
+            if red:
+                stdout += red
+            else:
+                # print(f'break')
+                break
+    except asyncio.TimeoutError:
+        # print('readline timeout')
+        pass
+
+    try:
+        while True:
+            red = await asyncio.wait_for(process.stdout.read(1), timeout=0.1)
+            # print(f'read {red=}')
+            if red:
+                stdout += red
+            else:
+                # print(f'break')
+                break
+    except asyncio.TimeoutError:
+        # print('read timeout')
+        pass
+
+    try:
+        body = stdout.decode()
+    except UnicodeDecodeError as e:
+        body = stdout[:e.start]
+    result.stdout = body
+    return result
 
 
-def terminate_running_processes(running_processes: list[subprocess.Popen], depth: int = 0) -> None:
+async def terminate_running_processes(running_processes: list[asyncio.subprocess.Process], depth: int = 0) -> None:
     for process in running_processes:
         print(' ' * depth + f'terminate {process}')
-        if process.poll() is None:
-            process.terminate()
-            try:
-                process.wait(timeout=5)
-            except subprocess.TimeoutExpired:
-                process.kill()
+        process.terminate()
+        try:
+            await asyncio.wait_for(process.wait(), timeout=1)
+        except asyncio.TimeoutError:
+            process.kill()
 
 
 def preview(value: Any, max_len: int = 64, separator: str = '...') -> str:
@@ -64,26 +98,28 @@ class Check:
     expected_format: Literal['text', 'json'] | None = None
     expected: str | None = None
 
-    def evaluate(self, depth: int = 0) -> subprocess.Popen | None:
+    async def evaluate(self, depth: int = 0) -> asyncio.subprocess.Process | None:
         print(' ' * depth + f"Run {self.language} code: {preview(self.code)}")
-        process_or_result = execute_code_block(self.code, language=self.language)
-        if isinstance(process_or_result, subprocess.Popen):
+        result: RunResult = await execute_code_block(self.code, language=self.language)
+        if result.process is not None:
             print(' ' * depth + f"Timeout {self.language} code, detach from it and kill later")
-            assert self.expected is None, f"{self.code=} got unexpected timeout"
-            return process_or_result
-        result = process_or_result
         expected = self.expected
         actual = result.stdout
         print(' ' * depth + f'Got {preview(actual)}... ', end='')
         if self.expected_format == 'json':
             expected = json.loads(expected)
             actual = json.loads(actual)
+        else:
+            expected = expected.strip()
+            actual = actual.strip()
         if expected:
             assert actual == expected, f"{self.code=} result does not match expected: {actual!r} != {expected!r}"
             print('OK')
         else:
             print('IGNORE')
 
+        return result.process
+
 
 @dataclass
 class LogicBlock:
@@ -93,16 +129,16 @@ class LogicBlock:
     def add(self, block: 'Check | LogicBlock') -> None:
         self.content.append(block)
 
-    def evaluate(self, depth: int = 0) -> None:
+    async def evaluate(self, depth: int = 0) -> None:
         print(' ' * depth + f"Starting logic block {self.name!r}")
         running_processes = []
         try:
             for block in self.content:
-                result = block.evaluate(depth + 1)
-                if isinstance(result, subprocess.Popen):
+                result = await block.evaluate(depth + 1)
+                if result is not None:
                     running_processes.append(result)
         finally:
-            terminate_running_processes(running_processes, depth)
+            await terminate_running_processes(running_processes, depth)
 
 
 def get_logic_block_from_md_file(file_path: str) -> LogicBlock:
